/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaDTree.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Decision tree.
**
*****************************************************************************/

#include "svlCudaDTree.h"

extern svlCudaMemHandler *cm;
typedef unsigned char uchar;

//void printTree(uchar *ptr)
//{
//	printf("%d %0.4f %d %0.4f %0.4f %0.4f %d %0.4f %0.4f %0.4f %0.4f\n",
//		*((int*)ptr), *((float*)ptr + sizeof(int)),
//		*((int*)ptr + sizeof(int) + sizeof(float)), *((float*)ptr + sizeof(int) + 2*sizeof(float)), *((float*)ptr + 16), *((float*)ptr + 20),
//		*((int*)ptr + 24), *((float*)ptr + 28), *((float*)ptr + 32), *((float*)ptr + 36), *((float*)ptr + 40));
//}

void printTree(int *iptr, float *fptr)
{
	/*printf("%d %0.4f %d %0.4f %0.4f %0.4f %d %0.4f %0.4f %0.4f %0.4f\n",
		*iptr, *fptr, *(iptr+1), *(fptr+1), *(fptr+1), *(fptr+1),
		*(iptr+1), *(fptr+1), *(fptr+1), *(fptr+1), *(fptr+1));*/
	printf("%d %d %d    %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f\n",
		*iptr, *(iptr+1), *(iptr+2),
		*fptr, *(fptr+2), *(fptr+3), *(fptr+4), *(fptr+5), *(fptr+6), *(fptr+7), *(fptr+8));
}

// Read in model from file.
svlCudaDTree::svlCudaDTree(const char *file)
{
	FILE *fp = fopen(file,"rb");
	if ( !fp ) {
		printf("Initialized svlCudaDTree with NULL FILE: %s.\n", file);
		return;
	}
	cout << "Going ...\n";

	// Read tree description in directly from file.
	fread(&n_trees,sizeof(int),1,fp);
	int *itree = (int*)malloc(sizeof(int)*n_trees*11);
	float *ftree = (float*)malloc(sizeof(float)*n_trees*11);
	int gi, *iptr = itree;
	float gf, *fptr = ftree;
	for ( int i=0; i<n_trees; ++i ) {
		fread(&gi,sizeof(int),1,fp); *iptr++ = gi;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gi,sizeof(int),1,fp); *iptr++ = gi;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gi,sizeof(int),1,fp); *iptr++ = gi;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
		fread(&gf,sizeof(float),1,fp); *fptr++ = gf;
	}
	fclose(fp);

	// Read in total tree.
	printf("%d %d\n", sizeof(int), sizeof(float));
	tree = (float*)malloc(sizeof(float)*n_trees*11);
	fp = fopen(file,"rb");
	fread(&gi,sizeof(int),1,fp);
	fread(tree,sizeof(float),n_trees*11,fp);
	fclose(fp);

	// Debug print
	/*for ( int i=0; i<n_trees; ++i )
		printTree(itree + i*3, ftree + i*8);*/

	// Allocate device memory and transfer in.
	dtree = cm->allocLine(n_trees*11);
	cm->lineToGPU(dtree, tree);

	//cm->line2file("line.out", dtree);
	
	free(itree);
	free(ftree);
}

svlCudaDTree::~svlCudaDTree()
{
	// Free host memory.s
	free(tree);

	// Free all device memory.
	/*for ( vector<svlCudaLine*>::iterator it = lines.begin(); it != lines.end(); ++it )
		cm->freeLine(*it);*/
	cm->freeLine(dtree);
}

