//todo-- add full lasik header
//Author: Ian Goodfellow

#include "svlAbstractCloud.h"
#include "svlPointColourer.h"


void svlAbstractCloud::preprocess(svlRenderType renderType) const
{
  bool point = false;
  bool surface = false;

  switch (renderType)
    {
    case SVL_RENDER_POINTS:
      point = true;
      break;

    case SVL_RENDER_STRUCTURED_SURFACE:
      surface = true;
      break;

    case SVL_RENDER_STRUCTURED_SURFACE_PRETTY:
      point = surface = true;
      break;

    default:
      SVL_LOG(SVL_LOG_FATAL, "svlAbstractCloud::preprocess was passed an unrecognized svlRenderType");
    }

  for (vector<const svlPointColourer *>::const_iterator i = universalCallbacks.begin(), last = universalCallbacks.end(); i != last; ++i)
    (*i)->preprocess(this);

  if (point)
    {
      for (vector<const svlPointColourer *>::const_iterator i = pointCallbacks.begin(), last = pointCallbacks.end(); i != last; ++i)
	(*i)->preprocess(this);
    }
  
  if (surface)
    {
      for (vector<const svlPointColourer *>::const_iterator i = surfaceCallbacks.begin(), last = surfaceCallbacks.end(); i != last; ++i)
    (*i)->preprocess(this);
    }
}

void svlAbstractCloud::process(const svlPoint3d & pt, bool usePointCallbacks, bool useSurfaceCallbacks) const
{
  //cout << "PROCESS CALLED" << endl;

  for (vector<const svlPointColourer *>::const_iterator i = universalCallbacks.begin(), last = universalCallbacks.end(); i != last; ++i)
    (*i)->colour(pt, this);

  if (usePointCallbacks)
    {
      for (vector<const svlPointColourer *>::const_iterator i = pointCallbacks.begin(), last = pointCallbacks.end(); i != last; ++i)
	(*i)->colour(pt, this);
    }
  
  if (useSurfaceCallbacks)
    {
      for (vector<const svlPointColourer *>::const_iterator i = surfaceCallbacks.begin(), last = surfaceCallbacks.end(); i != last; ++i)
	(*i)->colour(pt, this);
    }
}

svlAbstractCloud::~svlAbstractCloud()
{
}

svlAbstractCloud::svlAbstractCloud()
{
  cloudUpdated = true;
  assumeCloudAlwaysUpdated = false;
}
