/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCluster3d.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**   Generic clustering class. All implementations assume that asgs has been
**   allocted beforehand but not R_ix (which much be filled by the clustering
**   algorithm).
**
*****************************************************************************/

#include "svlCluster3d.h"

int svlCluster3d::relabelIx(vector<int> &ix) {
	// Relabel the unique labels in asgs sequentially from the start.
	// Return number of unique labels.
	int _max = -1;
	for ( vector<int>::const_iterator i=ix.begin(); i!=ix.end(); ++i ) {
		// find max label
		_max = _max > *i ? _max : *i;
	}
	vector<int> unique(_max+1,-1); // for mapping to unique addresses
	int lab = 0; // new assignment label
	for ( vector<int>::iterator i=ix.begin(); i!=ix.end(); ++i ) {
		if ( unique[*i] == -1 ) { // new label
			unique[*i] = lab;
			++lab;
		}
		*i = unique[*i];
	}
	return _max+1;
}

void svlCluster3d::populateIx(vector<vector<int> > &regionIx, vector<int> &ix) {
#if 1
	int j=0;
	for ( vector<vector<int> >::const_iterator j1=regionIx.begin(); j1!=regionIx.end(); ++j1 ) {
		for ( vector<int>::const_iterator k1=(*j1).begin(); k1!=(*j1).end(); ++k1 ) {
			ix[*k1] = j;
		}
		++j;
	}
#else
	for ( int j=0; j<(int)regionIx.size(); ++j ) {
		for ( int i=0; i<(int)regionIx[j].size(); ++i ) {
			ix[ regionIx[j][i] ] = j;
		}
	}
#endif
}

void svlCluster3d::populateRegionIx(vector<int> &ix, int n_max, vector<vector<int> > &regionIx) {
	// Convert point-to-assignment vector into cluster-to-points R_ix.
	regionIx = vector<vector<int> >(n_max,vector<int>());
	for ( int i=0; i<(int)ix.size(); ++i ) {
		if ( ix[i] == -1 )
			continue;
		regionIx[ix[i]].push_back(i);
	}
}

void svlCluster3d::populateRegSizes(vector<vector<int> > &regionIx, svlSortVector<int,int> &regionSizes) {
	regionSizes.clear();
	regionSizes.reserve((int)regionIx.size());
	for ( int i=0; i<(int)regionIx.size(); ++i ) {
		regionSizes.push_back((int)regionIx[i].size());
	}
}

void svlCluster3d::populateRegionIxAndSizes(vector<int> &ix, int n_max, vector<vector<int> > &regionIx, svlSortVector<int,int> &regionSizes) {
	populateRegionIx(ix,n_max,regionIx);
	populateRegSizes(regionIx,regionSizes);
}

