/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlKMeansClustering.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
*****************************************************************************/

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#endif

#include "svlKMeansClustering.h"

#define TIME_BOOL1 0
// TODO: Turn conv_thresh into one of those parameters
#define DIST_METHOD_1 0 // || p - p' || + || n - n' ||
#define DIST_METHOD_2 1 // | (p - p').n' | - a * | n.n |
#define DEBUG_OUT 0

svlKMeansClustering::svlKMeansClustering() {
	declareOption("EmptyAction","Restart"); // when a cluster is empty, restart it
	declareOption("Centers","Initialize"); // when we start k-means, initialize the centers to random
	declareOption("MaxIter","100");
	declareOption("ConvThresh","0.003");
	declareOption("Merge","true");
	declareOption("% of max # pts","0.0");
}

svlKMeansClustering::~svlKMeansClustering() {
}

void svlKMeansClustering::setDefaultOptions() {
	setOption("EmptyAction","Restart");
	setOption("Centers","Initialize");
	setOption("MaxIter","100");
	setOption("% of max # pts","0.0");
}

void svlKMeansClustering::cluster(svlPointCloudData &PCD, vector<int> &asgs, vector<vector<int> > &R_ix, vector<svlPlane> &planes, int k) {
	do_kmeans(PCD.pointCloud, PCD.pointNormals, asgs, R_ix, planes, k);
}
/*
void svlKMeansClustering::do_kmeans(const vector<svlPoint3d> &Pts, const vector<svlPoint3d> &N, vector<int> &asgns, vector<vector<int> > &R_ix, vector<svlCentroid6> &cents, int k) {
	// perform k-means on the points and normals
	assert(k>0);
	assert(Pts.size()==N.size());

	// initializations
	int *ct = new int[k], n_pts = (int)Pts.size(), gi, iter, ii;
	vector<svlCentroid6> centsp(k,svlCentroid6()); // centers' (prime, or old)
	double *dists = new double[k*n_pts];
	bool *center_enabled = new bool[k];
	for ( int i=0; i<k; ++i ) {
		center_enabled[i] = true;
	}
	vector<svlPoint3d>::const_iterator ip, in; // iterators over the point vectors
	double alpha = 1.0; // regularization on the normal distance
	double convthresh = getOptionAsDouble("ConvThresh") * (double)k;

#if DIST_METHOD_2
	// allocate matrices for inversion
	//CvMat* PPT = cvCreateMat(3,3,CV_32FC1); // (P-C)*(P-C)^T
    CvMat* U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat** PPTS = (CvMat**)malloc(sizeof(U)*k);
	for ( int i=0; i<k; ++i )
		PPTS[i] = cvCreateMat(3,3,CV_32FC1);
    CvMat* S = cvCreateMat(3,3,CV_32FC1); // singular values
#endif
	
	// initialize cluster centroids to random points
	srand ( (unsigned int)time(NULL) );
	if ( strcmp(getOption("Centers").c_str(),"Initialize") == 0 ) {
		for ( int i=0; i<k; ++i ) {
			gi = rand()%n_pts;
			cents[i].p = Pts[gi];
			cents[i].n = N[gi];
			centsp[i].p = svlPoint3d(-1e3);
			centsp[i].n = svlPoint3d(-1e3);
		}
	}
	// else, keep them as they've been fed in, assuming they have been initialized.

	// convergence distance sums
	double gd = 0.0;
	for ( int i=0; i<k; ++i )
		gd += sqrt( (cents[i].p-centsp[i].p).norm2() + (cents[i].n-centsp[i].n).norm2() );

	iter = 0;
	int max_iter = getOptionAsInt("MaxIter");
	while ( gd >= convthresh && ++iter <= max_iter ) {
#if TIME_BOOL1
	clock_t histc1 = clock();
	double hister1 = 1000;
	for ( int jj1=0; jj1<(int)(hister1); ++jj1 ) {
#endif
		// calculate distances to centers
		ip = Pts.begin();
		in = N.begin();
		ii = -1;
		while ( ++ii < n_pts ) { // outer loop over all points
			for ( int j=0; j<k; ++j ) // inner loop over centroids
				if ( center_enabled[j] )
#if DIST_METHOD_1
					dists[j*n_pts + ii] = (*ip-cents[j].p).norm2() + (*in-cents[j].n).norm2();
#elif DIST_METHOD_2
					dists[j*n_pts + ii] = fabs( (*ip-cents[j].p).dot(cents[j].n) );// - alpha*fabs( in->dot(cents[j].n) );
					//dists[j*n_pts + ii] = fabs( (*ip).dot(cents[j].n) - cents[j].p );// - alpha*fabs( in->dot(cents[j].n) );
#endif
			++ip;
			++in;
		}
#if TIME_BOOL1
	}
	cout << "Dist to centers (" << hister1 << "/per) -- " << (((double)clock()-histc1)/CLOCKS_PER_SEC)/hister1 << endl;
	histc1 = clock();
	hister1 = 1000;
	for ( int jj1=0; jj1<(int)(hister1); ++jj1 ) {
#endif
		// find minimum dist for each point
		for ( int i=0; i<n_pts; ++i ) {
			gd = 1e9;
			for ( int j=0; j<k; ++j ) {
				if ( center_enabled[j] ) {
					if ( dists[j*n_pts + i] < gd ) {
						gd = dists[j*n_pts + i];
						asgns[i] = j;
					}
				}	
			}
		}
#if TIME_BOOL1
	}
	cout << "Min dist (" << hister1 << "/per) -- " << (((double)clock()-histc1)/CLOCKS_PER_SEC)/hister1 << endl;
	histc1 = clock();
	hister1 = 1000;
	for ( int jj1=0; jj1<(int)(hister1); ++jj1 ) {
#endif
		// shift-&-zeralize centers
		for ( int j=0; j<k; ++j ) {
			if ( center_enabled[j] ) {
				centsp[j].p = cents[j].p;
				centsp[j].n = cents[j].n; // move current down to p
				cents[j].p.set(0.0);
				cents[j].n.set(0.0); // init these as accumulators
			}
		}
		// recalculate centers
#if DIST_METHOD_1
		ip = Pts.begin();
		in = N.begin();
		int ix;
		for ( int j=0; j<k; ++j ) ct[j] = 0; // zeralize counts
		ii = -1;
		while ( ++ii < n_pts ) {
			ix = asgns[ii];
			cents[ix].p += *ip;
			cents[ix].n += *in;
			++ip;
			++in; // pre is faster than post. A lot in this case. svlPoint3d's and all ...
			++ct[ix];
		}
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				cents[j].p /= (double)ct[j];
				cents[j].n /= (double)ct[j];
			}
		}
#elif DIST_METHOD_2
		ip = Pts.begin();
		int ix;
		for ( int j=0; j<k; ++j ) ct[j] = 0; // zeralize counts
		ii = -1;
		while ( ++ii < n_pts ) {
			ix = asgns[ii];
			cents[ix].p += *ip;
			++ct[ix];
			++ip;
		}
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				cents[j].p /= (double)ct[j]; // calculate assigned-points centroid
				cvZero(PPTS[j]);
#if DEBUG_OUT
				cout << "P: " << cents[j].p << endl;
				cout << j << ": " << CV_MAT_ELEM(*PPTS[j],float,0,0)
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,1)
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,2) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,1) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,2) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,2,2) << endl;
#endif
			}
		}
		ip = Pts.begin();
		ii = -1;
		while ( ++ii < n_pts ) {
			svlPoint3d pt = *ip - cents[asgns[ii]].p;
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,0) += (float)(pt.x*pt.x);
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,1) += (float)(pt.x*pt.y);
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,2) += (float)(pt.x*pt.z);
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,1,1) += (float)(pt.y*pt.y);
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,1,2) += (float)(pt.y*pt.z);
			CV_MAT_ELEM(*PPTS[asgns[ii]],float,2,2) += (float)(pt.z*pt.z);
			++ip;
		}
		// set lower-triangular part of covariance matrix
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				CV_MAT_ELEM(*PPTS[j],float,1,0) = CV_MAT_ELEM(*PPTS[j],float,0,1);
				CV_MAT_ELEM(*PPTS[j],float,2,0) = CV_MAT_ELEM(*PPTS[j],float,0,2);
				CV_MAT_ELEM(*PPTS[j],float,2,1) = CV_MAT_ELEM(*PPTS[j],float,1,2);
				cvSVD(PPTS[j],S,U,NULL,0); // SVD
				cents[j].n.x = CV_MAT_ELEM(*U,float,0,2);
				cents[j].n.y = CV_MAT_ELEM(*U,float,1,2);
				cents[j].n.z = CV_MAT_ELEM(*U,float,2,2);
#if DEBUG_OUT
				cout << j << ": " << CV_MAT_ELEM(*PPTS[j],float,0,0)
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,1)
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,2) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,1) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,2) 
					 << ", " << CV_MAT_ELEM(*PPTS[j],float,2,2) << endl;
				cout << "New normal: " << cents[j].n << endl;
#endif
			}
		}
#endif
		// sum deviations from old centers
		gd = 0.0;
		for ( int j=0; j<k; ++j ) {
			if ( center_enabled[j] ) {
				if ( ct[j] == 0 ) {
					if ( strcmp(getOption("EmptyAction").c_str(),"Restart") == 0 ) { // restart center at random
						gi = rand()%n_pts;
						cents[j].p = Pts.at(gi);
						cents[j].n = N.at(gi);
					} else if ( strcmp(getOption("EmptyAction").c_str(),"Drop") == 0 ) { // drop center from rest of algorithm
						center_enabled[j] = false; // stop tracking and updating centroid j
					}
				}
				gd += sqrt( (cents[j].p-centsp[j].p).norm2() + (cents[j].n-centsp[j].n).norm2() ); // add distance
			}
		}
#if TIME_BOOL1
	}
	cout << "Recalc centers (" << hister1 << "/per) -- " << (((double)clock()-histc1)/CLOCKS_PER_SEC)/hister1 << endl;
#endif
		//cout << "Iter # " << iter << " -- Conv: " << gd << endl;
	}

	//int n_max = relabel_asgs(asgns); // cannot do this or assignments won't agree with centroids in CENTS

#if DIST_METHOD_2
	for ( int i=0; i<k; ++i )
		cvReleaseMat(&PPTS[i]);
    cvReleaseMat(&U);
    cvReleaseMat(&S);
#endif

#if 0
	FILE *fp = fopen("k_n.out","wb");
	fwrite(&n_pts,sizeof(int),1,fp);
	for ( int i=0; i<n_pts; ++i ) {
		fwrite(&Pts[i].x,sizeof(double),1,fp);
		fwrite(&Pts[i].y,sizeof(double),1,fp);
		fwrite(&Pts[i].z,sizeof(double),1,fp);
	}
	for ( int i=0; i<k; ++i ) {
		if ( center_enabled[i] ) {
			fwrite(&cents[i].p.x,sizeof(double),1,fp);
			fwrite(&cents[i].p.y,sizeof(double),1,fp);
			fwrite(&cents[i].p.z,sizeof(double),1,fp);
			fwrite(&cents[i].n.x,sizeof(double),1,fp);
			fwrite(&cents[i].n.y,sizeof(double),1,fp);
			fwrite(&cents[i].n.z,sizeof(double),1,fp);
		}
	}
	fclose(fp);
#endif

	// MergeClusters
	if ( getOptionAsBool("Merge") ) {
		for ( int i=0; i<k; ++i ) {
			if ( center_enabled[i] ) {
				for ( int j=i+1; j<k; ++j ) {
					if ( center_enabled[j] ) {
						svlPoint3d p_offset = cents[i].p - cents[j].p;
						if ( fabs(cents[i].n.dot(cents[j].n)) >= 0.93
							&& fabs(p_offset.dot(cents[i].n)) <= 0.5
							&& fabs(p_offset.dot(cents[j].n)) <= 0.5 ) {
							// merge cluster j into cluster i
							for ( int k=0; k<n_pts; ++k ) {
								if ( asgns[k] == j ) {
									asgns[k] = i;
								}
							}
							center_enabled[j] = false;
						}
					}
				}
			}
		}
	}

	// TODO [PB]: Have relabel_asgs return a permutation vector.
	populate_Rix(asgns,k,R_ix);

	int writei;
	double writed = getOptionAsDouble("% of max # pts");
	FILE *fp = fopen("km.out","wb");
	fwrite(&writed,sizeof(double),1,fp);
	if ( (gd=getOptionAsDouble("% of max # pts")) > 0.0 ) {
		// use only major clusters; dropping remainder
		svlIxVector<int,int> regionSizes((int)R_ix.size(),true); // true reverses sort (descending)
		for ( int i=0; i<(int)R_ix.size(); ++i ) {
			regionSizes.push_back((int)R_ix[i].size());
		}
		regionSizes.sort();
		int i = 0;
		// count off valid clusters
		for ( ; i<(int)regionSizes.size(); ++i ) {
			if ( regionSizes.at(i) < gd*regionSizes.at(0) )
				break;
		}
		// falsilize remaining centers
		for ( int j=i; j<(int)regionSizes.size(); ++j ) {
			writei = regionSizes.at(j);
			fwrite(&writei,sizeof(int),1,fp);
			center_enabled[regionSizes.ix_at(j)] = false;
		}
		// reassign any points that were before assigned to the newly-falsilized clusters
		for ( int j=i; j<(int)regionSizes.size(); ++j ) {
			for ( vector<int>::const_iterator n=R_ix[regionSizes.ix_at(j)].begin(); n!=R_ix[regionSizes.ix_at(j)].end(); ++n ) {
				// find new min for this point *n
				gd = 1e9;
				for ( int m=0; m<k; ++m ) {
					if ( center_enabled[m] ) {
						if ( dists[m*n_pts + i] < gd ) {
							gd = dists[m*n_pts + *n];
							asgns[*n] = m;
						}
					}	
				}
				// manage R_ix
				R_ix[asgns[*n]].push_back(*n);
			}
			// delete the now-empty region indices vector for defunct cluster
			R_ix[regionSizes.ix_at(j)].clear();
		}
	}
	fclose(fp);

	delete[] ct;
	delete[] dists;
	delete[] center_enabled;
}

void svlKMeansClustering::do_kmeans_acc(const vector<svlPoint3d> &Pts, const vector<svlPoint3d> &N, vector<svlCentroid6> &cents, vector<int> &asgns, int k) {
	// perform accelerated k-means on the data
	assert(k>0);
	assert(Pts.size()==N.size());

	// initializations
	int *ct = new int[k], n_pts = (int)Pts.size(), gi, ii;
	double gd, gd1;
	vector<svlCentroid6> centsp(k,svlCentroid6()); // centers and centers' (prime, or old)
	double *dists = new double[k*n_pts], // point-to-centroid distances
		   *low = new double[k*n_pts], // lower bounds
		   *upp = new double[n_pts], // upper bounds
		   *d_cc = new double[k*k], // centroid-to-centroid distances
		   *d_ccp = new double[k], // centroid-to-centroid' distances
		   *s = new double[k]; // c-c' measure
	bool *r = new bool[n_pts];
	vector<svlPoint3d>::const_iterator ip, in; // iterators over the point vectors
	
	// constant initializations
	for ( int i=0; i<k*n_pts; ++i ) {
		low[i] = 0.0;
		upp[i] = 0.0;
		dists[i] = 1e9; // superlatively large number
	}
	// initialize cluster centroids to random points
	srand ( (unsigned int)time(NULL) );
	for ( int i=0; i<k; ++i ) {
		gi = rand()%n_pts;
		cents[i].p = Pts[gi];
		cents[i].n = N[gi];
		centsp[i].p = svlPoint3d(-1e3);
		centsp[i].n = svlPoint3d(-1e3);
		//cout << gi << ": " << cents[i].p << " ,, " << cents[i].n << endl;
	}
	for ( int i=0; i<k; ++i ) { // inter-centroid distances
		for ( int j=i+1; j<k; ++j ) {
			d_cc[i*k + j] = (cents[i].p-cents[j].p).norm2() + (cents[i].n-cents[j].n).norm2();
			d_cc[j*k + i] = d_cc[i*k + j];
		}
	}

	// 0. Initial point-to-centroid assignments
	ip = Pts.begin();
	in = N.begin();
	ii = -1;
	while ( ++ii < n_pts ) { // loop x
		gd = 1e9; // for min_c d(x,c)
		for ( int j=0; j<k; ++j ) { // loop c
			if ( j == 0 ) { // need at least one distance calculation
				gd1 = (*ip-cents[j].p).norm2() + (*in-cents[j].n).norm2();
				dists[j*n_pts + ii] = gd1;
				low[j*n_pts + ii] = gd1; // set l(x,c)
				if ( gd1 < gd ) {
					gd = gd1;
					asgns[ii] = j;
				}
			} else { // Lemma 1
				for ( int l=0; l<k; ++l ) {
					if ( d_cc[j*k + l] < 2*dists[l*n_pts + ii] ) { // d(b,c) < 2d(x,b), must calculate d(x,c)
						gd1 = (*ip-cents[j].p).norm2() + (*in-cents[j].n).norm2();
						dists[j*n_pts + ii] = gd1;
						low[j*n_pts + ii] = gd1; // set l(x,c)
						if ( gd1 < gd ) {
							gd = gd1;
							asgns[ii] = j;
						}
					}
				}
			}
		}
		upp[ii] = gd;
		++ip;
		++in;
	}
	
	gd = 1e9;
	while ( gd > (double)k ) {
		// Next, repeat until convergence:

		// 1. For all center c and c', compute d(c,c'). For all c, compute s(c) = 1/2 min_{c'!=c} d(c,c')
		for ( int i=0; i<k; ++i ) { // inter-centroid distances
			for ( int j=i+1; j<k; ++j ) {
				d_cc[i*k + j] = (cents[i].p-cents[j].p).norm2() + (cents[i].n-cents[j].n).norm2();
				d_cc[j*k + i] = d_cc[i*k + j];
			}
		}
		for ( int i=0; i<k; ++i ) { // inter-centroid distances
			gd = 1e9;
			for ( int j=0; j<k; ++j ) {
				if ( i == j ) continue;
				if ( d_cc[i*k + j] < gd )
					gd = d_cc[i*k + j];
			}
			s[i] = 0.5*gd;
		}

		// 2. Identify all points x such that u(x) <= s(c(x)).
		// 3. For all remaining points x and centers c such that
		//    (i) c != c(x) and
		//	  (ii) u(x) > l(x,c) and
		//    (iii) u(x) > 1/2 d(c(x),x), then ...
		ip = Pts.begin();
		in = N.begin();
		ii = -1;
		while ( ++ii < n_pts ) {
			if ( upp[ii] > s[asgns[ii]] ) {
				for ( int j=0; j<k; ++j ) {
					if ( j == asgns[ii]
						 && upp[ii] < low[j*n_pts + ii]
						 && upp[ii] < 0.5*d_cc[j*k + asgns[ii]] ) continue;
					if ( r[ii] ) { // compute d(x,c(x))
						dists[asgns[ii]*n_pts + ii] = (*ip-cents[asgns[ii]].p).norm2() + (*in-cents[asgns[ii]].n).norm2();
						r[ii] = false;
					} else {
						dists[asgns[ii]*n_pts + ii] = upp[ii];
					}
					if ( dists[asgns[ii]*n_pts + ii] > low[j*n_pts + ii]
						 || dists[asgns[ii]*n_pts + ii] > 0.5*d_cc[j*k + asgns[ii]] ) {
						dists[j*n_pts + ii] = (*ip-cents[j].p).norm2() + (*in-cents[j].n).norm2();
						if ( dists[j*n_pts + ii] < dists[asgns[ii]*n_pts + ii] )
							asgns[ii] = j;
					}
				}
			}
			++ip;
			++in;
		}

		// 4. For each center c, let m(c) be the mean of the points assigned to c.
		ip = Pts.begin();
		in = N.begin();
		ii = -1;
		for ( int j=0; j<k; ++j ) {
			centsp[j].p.set(0.0);
			centsp[j].n.set(0.0);
			ct[j] = 0;
		}
		while ( ++ii < n_pts ) {
			centsp[asgns[ii]].p += *ip;
			centsp[asgns[ii]].n += *ip;
			++ct[asgns[ii]];
			++ip;
			++in;
		}
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				centsp[j].p /= (double)ct[j];
				centsp[j].n /= (double)ct[j];
			}
		}

		// 5. For each point x and center c, assign l(x,c) = max(0, l(x,c)-d(c,m(c))).
		gd = 0.0;
		for ( int i=0; i<k; ++i ) { // centroid-to-centroid' distances
			d_ccp[i] = (cents[i].p-centsp[i].p).norm2() + (cents[i].n-centsp[i].n).norm2();
			gd += d_ccp[i];
		}
		for ( int i=0; i<n_pts; ++i )
			for ( int j=0; j<k; ++j )
				low[j*n_pts + i] = max( 0.0, low[j*n_pts + i] - d_cc[j] );

		// 6. For each point x, assign u(x) = u(x) + d(m(c(x)),c(x)); r(x) = true;
		ii = -1;
		while ( ++ii < n_pts ) {
			upp[ii] += d_ccp[asgns[ii]];
			r[ii] = true;
		}

		// 7. Replace each center c by m(c).
		for ( int j=0; j<k; ++j ) {
			cents[j].p = centsp[j].p;
			cents[j].n = centsp[j].n;
		}
		cout << "Iter: " << gd << endl;
	}

	delete[] ct;
	delete[] dists;
	delete[] low;
	delete[] upp;
	delete[] d_cc;
	delete[] d_ccp;
	delete[] s;
	delete[] r;
}
*/

//void svlKMeansClustering::do_kmeans(const vector<svlPoint3d> &Pts, const vector<svlPoint3d> &N, vector<int> &asgns, vector<vector<int> > &R_ix, list<svlCentroid6> &cents, int k) {
//	// perform k-means on the points and normals
//	assert(k>0);
//	assert(Pts.size()==N.size());
//
//	// initializations
//	int *ct = new int[k], n_pts = (int)Pts.size(), gi, iter, ii, jj;
//	list<svlCentroid6> centsp(k,svlCentroid6(-1e6)); // centers' (prime, or old)
//	double *dists = new double[k*n_pts];
//
//	vector<svlPoint3d>::const_iterator ip, in; // iterators over the point vectors
//	double alpha = 1.0; // regularization on the normal distance
//	double convthresh = getOptionAsDouble("ConvThresh") * (double)k;
//
//#if DIST_METHOD_2
//	// allocate matrices for inversion
//	//CvMat* PPT = cvCreateMat(3,3,CV_32FC1); // (P-C)*(P-C)^T
//    CvMat* U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
//	CvMat** PPTS = (CvMat**)malloc(sizeof(U)*k); // one covariance matrix for each center
//				// this way we only need to loop through the points once
//	for ( int i=0; i<k; ++i )
//		PPTS[i] = cvCreateMat(3,3,CV_32FC1);
//    CvMat* S = cvCreateMat(3,3,CV_32FC1); // singular values
//#endif
//	
//	// creating missing centroids; initialize to random
//	srand ( (unsigned int)time(NULL) );
//	if ( (int)cents.size() < k ) { // initialize new, random centroids
//		for ( int i=(int)cents.size(); i<k; ++i ) {
//			//if ( strcmp(getOption("Centers").c_str(),"Initialize") == 0 ) {
//			gi = rand()%n_pts;
//			cents.push_back(svlCentroid6(Pts[gi],N[gi]));
//		}
//	}
//	// else, keep them as they've been fed in, assuming they have been initialized.
//
//	// convergence distance sums
//	double gd = 0.0;
//	list<svlCentroid6>::iterator c_it, cp_it; // centroid list iterators
//	c_it = cents.begin();
//	cp_it = centsp.begin();
//	jj = -1;
//	while ( ++jj < k ) {
//		gd += sqrt( (*c_it - *cp_it).norm2() );
//		++c_it;
//		++cp_it;
//	}
//
//	iter = 0;
//	int max_iter = getOptionAsInt("MaxIter");
//	while ( gd >= convthresh && ++iter <= max_iter ) {
//		// calculate distances to centers
//		ip = Pts.begin();
//		in = N.begin();
//		ii = -1;
//		while ( ++ii < n_pts ) { // outer loop over all points
//			c_it = cents.begin();
//			jj = -1;
//			while ( ++jj < k ) { // inner loop over centroids
//#if DIST_METHOD_1
//				dists[jj*n_pts + ii] = (*ip - c_it->p ).norm2() + (*in - c_it->n ).norm2();
//#elif DIST_METHOD_2
//				dists[jj*n_pts + ii] = abs( (*ip - c_it->p ).dot(c_it->n) ) - alpha*abs( in->dot(c_it->n) );
//#endif
//				++c_it;
//			}
//			++ip;
//			++in;
//		}
//		// find minimum dist for each point
//		for ( ii=0; ii<n_pts; ++ii ) {
//			gd = 1e9;
//			c_it = cents.begin();
//			jj = -1;
//			while ( ++jj < k ) {
//				if ( dists[jj*n_pts + ii] < gd ) {
//					gd = dists[jj*n_pts + i];
//					asgns[ii] = jj;
//				}
//			}
//		}
//		// shift-&-zeralize centers
//		c_it = cents.begin();
//		cp_it = cents[.begin();
//		jj = -1;
//		while ( ++jj < k ) {
//			*cp_it = *c_it; // store current centroids
//			c_it->set(0.0); // init these as accumulators
//			++c_it;
//			++cp_it;
//		}
//		// recalculate centroids
//#if DIST_METHOD_1
//		ip = Pts.begin();
//		in = N.begin();
//		int ix;
//		for ( int j=0; j<k; ++j ) ct[j] = 0; // zeralize counts
//		ii = -1;
//		while ( ++ii < n_pts ) {
//			ix = asgns[ii];
//			cents[ix].p += *ip;
//			cents[ix].n += *in;
//			++ip;
//			++in; // pre is faster than post. A lot in this case. svlPoint3d's and all ...
//			++ct[ix];
//		}
//		c_it = cents.begin();
//		jj = -1;
//		while ( ++jj < k ) {
//			*c_it /= (double)ct[j];
//		}
//#elif DIST_METHOD_2
//		ip = Pts.begin();
//		int ix;
//		for ( int j=0; j<k; ++j ) ct[j] = 0; // zeralize counts
//		ii = -1;
//		while ( ++ii < n_pts ) {
//			ix = asgns[ii];
//			cents[ix].p += *ip;
//			++ct[ix];
//			++ip;
//		}
//		for ( int j=0; j<k; ++j ) {
//			if ( ct[j] > 0 ) {
//				cents[j].p /= (double)ct[j]; // calculate assigned-points centroid
//				cvZero(PPTS[j]);
//#if DEBUG_OUT
//				cout << "P: " << cents[j].p << endl;
//				cout << j << ": " << CV_MAT_ELEM(*PPTS[j],float,0,0)
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,1)
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,2) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,1) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,2) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,2,2) << endl;
//#endif
//			}
//		}
//		ip = Pts.begin();
//		ii = -1;
//		while ( ++ii < n_pts ) {
//			svlPoint3d pt = *ip - cents[asgns[ii]].p;
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,0) += (float)(pt.x*pt.x);
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,1) += (float)(pt.x*pt.y);
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,0,2) += (float)(pt.x*pt.z);
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,1,1) += (float)(pt.y*pt.y);
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,1,2) += (float)(pt.y*pt.z);
//			CV_MAT_ELEM(*PPTS[asgns[ii]],float,2,2) += (float)(pt.z*pt.z);
//			++ip;
//		}
//		// set lower-triangular part of covariance matrix
//		for ( int j=0; j<k; ++j ) {
//			if ( ct[j] > 0 ) {
//				CV_MAT_ELEM(*PPTS[j],float,1,0) = CV_MAT_ELEM(*PPTS[j],float,0,1);
//				CV_MAT_ELEM(*PPTS[j],float,2,0) = CV_MAT_ELEM(*PPTS[j],float,0,2);
//				CV_MAT_ELEM(*PPTS[j],float,2,1) = CV_MAT_ELEM(*PPTS[j],float,1,2);
//				cvSVD(PPTS[j],S,U,NULL,0); // SVD
//				cents[j].n.x = CV_MAT_ELEM(*U,float,0,2);
//				cents[j].n.y = CV_MAT_ELEM(*U,float,1,2);
//				cents[j].n.z = CV_MAT_ELEM(*U,float,2,2);
//#if DEBUG_OUT
//				cout << j << ": " << CV_MAT_ELEM(*PPTS[j],float,0,0)
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,1)
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,0,2) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,1) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,1,2) 
//					 << ", " << CV_MAT_ELEM(*PPTS[j],float,2,2) << endl;
//				cout << "New normal: " << cents[j].n << endl;
//#endif
//			}
//		}
//#endif
//		// sum deviations from old centers
//		gd = 0.0;
//		for ( list<svlCentroid6>::iterator j=cents.begin(); j!=cents.end(); ) {
//			if ( ct[j] == 0 ) { // emtpy centroid
//				if ( strcmp(getOption("EmptyAction").c_str(),"Restart") == 0 ) { // restart center at random
//					gi = rand()%n_pts;
//					(*j).set(Pts[gi],N[gi]);
//					gd += sqrt( (cents[j]-centsp[j]).norm2() ); // add distance
//					++j; // advance
//				} else if ( strcmp(getOption("EmptyAction").c_str(),"Drop") == 0 ) { // drop center from rest of algorithm
//					j = cents.erase(j); // remove cluster; and advances pointer, too
//				}
//			} else {
//				gd += sqrt( (cents[j]-centsp[j]).norm2() ); // add distance
//				++j;
//			}
//		}
//		k = (int)cents.size();
//		//cout << "Iter # " << iter << " -- Conv: " << gd << endl;
//	}
//
//	//int n_max = relabel_asgs(asgns); // cannot do this or assignments won't agree with centroids in CENTS
//	// TODO [PB]: Have relabel_asgs return a permutation vector.
//	populate_Rix(asgns,k,R_ix);
//
//#if DIST_METHOD_2
//	for ( int i=0; i<k; ++i )
//		cvReleaseMat(&PPTS[i]);
//    cvReleaseMat(&U);
//    cvReleaseMat(&S);
//#endif
//
//	delete[] ct;
//	delete[] dists;
//	delete[] center_enabled;
//}

void svlKMeansClustering::do_kmeans(const vector<svlPoint3d> &P, const vector<svlPoint3d> &N, vector<int> &ix, vector<vector<int> > &rix, vector<svlPlane> &planes, int k) {
	// perform k-means on the points and normals
	assert(k>0);
	//planes.resize(k,svlPlane());

	// initializations
	int *ct = new int[k], n_pts = (int)P.size(), gi, iter, ii;

	vector<svlPlane> planesp(k,svlPlane()); // old planes
	vector<svlPoint3d>::const_iterator ip, in; // iterators over the point vectors
	double alpha = 1.0; // regularization on the normal distance
	double convthresh = getOptionAsDouble("ConvThresh") * (double)k;
	double *w = new double[n_pts], *ct_var = new double[k], *ct_mean = new double[k], *w_tot = new double[k];

	ix.resize(n_pts,-1);
	rix = vector<vector<int> >(k,vector<int>()); // init empty region indices double-vector

	// allocate matrices for multiple SVD operations
	//CvMat* PPT = cvCreateMat(3,3,CV_32FC1); // (P-C)*(P-C)^T
    CvMat* U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat** PPTS = (CvMat**)malloc(sizeof(U)*k);
	for ( int i=0; i<k; ++i )
		PPTS[i] = cvCreateMat(3,3,CV_32FC1);
    CvMat* S = cvCreateMat(3,3,CV_32FC1); // singular values
	
	// initialize cluster centroids to random points
	srand ( (unsigned int)time(NULL) );
	//if ( strcmp(getOption("Centers").c_str(),"Initialize") == 0 ) {
	for ( int i=(int)planes.size(); i<k; ++i ) { // add additional, random centroids to get up to quota
		gi = rand()%n_pts;
		planes.push_back(svlPlane());
		planes.back().p = P[gi];
#if 0
		planes.back().n = N[gi];
#else
		planes.back().n.x = (double)(rand() - (RAND_MAX>>1));
		planes.back().n.y = (double)(rand() - (RAND_MAX>>1));
		planes.back().n.z = (double)(rand() - (RAND_MAX>>1));
		planes.back().n /= planes[i].n.norm();
#endif
		planes.back().off = 0.0;
		planes.back().flag = true;
		planesp[i].p = svlPoint3d(-1e6); // planesp has already been allocated
		planesp[i].n = svlPoint3d(-1e6);
	}
	// else, keep them as they've been fed in, assuming they have been initialized.
	for ( int i=0; i<k; ++i ) { // enable all planes
		planes[i].flag = true;
	}

	// convergence distance sums
	double gd = 0.0, gd1;
	for ( int i=0; i<k; ++i ) {
		gd += sqrt( (planes[i].p-planesp[i].p).norm2() + (planes[i].n-planesp[i].n).norm2() );
	}

	iter = 0;
	int max_iter = getOptionAsInt("MaxIter"), asg;
	while ( gd >= convthresh && ++iter <= max_iter ) {
		// find minimum dist for each point, building asgns and rix
		for ( int j=0; j<k; ++j ) { // shift-&-zeralize centers
			ct[j] = 0;
			if ( planes[j].flag ) {
				planesp[j].set(planes[j]); // shift to old planes
				planes[j].set(0.0); // init planes as accumulators
				//rix[j].clear();
			}
		}
		ip = P.begin();
		in = N.begin();
		ii = 0;
		for ( ; ii<n_pts; ++ii, ++ip, ++in ) { // outer loop over all points
			asg = -1;
			gd = 1e9;
			for ( int j=0; j<k; ++j ) { // inner loop over centroids
				if ( planes[j].flag ) {
					//cout << "n: " << planes[j].n << " --- " << planes[j].n.dot(*in) << endl;
					if ( fabs(planesp[j].n.dot(*in)) >= 0.9 ) { // inf distance otherwise
						gd1 = planesp[j].n.dot(*ip) + planesp[j].off;// - alpha*fabs( in->dot(planes[j].n) );
						if ( fabs(gd1) < fabs(gd) ) {
							gd = gd1;
							asg = j;
						}
					}
				}
			}
			if ( asg == -1 )
				continue;
			// compute rix and update next plane centroids as we go
			ix[ii] = asg;
			planes[asg].p += *ip;
			planes[asg].n += *in;
			++ct[asg];
		}
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				planes[j].p /= (double)ct[j];
				planes[j].n /= planes[j].n.norm();
				planes[j].off = -planes[j].n.dot(planes[j].p);
			}
		}
#if 1
		for ( int ms=0; ms<2; ++ms ) {
			for ( int j=0; j<k; ++j )
				if ( ct[j] > 0 )
					w_tot[j] = ct_mean[j] = ct_var[j] = 0.0; // zeralize accumulators
			// mean-shift correct on (ip-p_c)^T n_c
			ii = 0;
			ip = P.begin();
			for ( ; ii<n_pts; ++ii, ++ip ) {
				if ( (asg=ix[ii]) == -1 ) continue;
				w[ii] = planes[asg].n.dot(*ip) + planes[asg].off;
				ct_mean[asg] += w[ii]; // \sum x
				ct_var[asg] += w[ii]*w[ii]; // \sum x^2
			}
			ii = 0;
			ip = P.begin();
			for ( int j=0; j<k; ++j ) { // compute stddev of norm dist
				if ( ct[j] == 0 ) continue;
				ct_mean[j] /= (double)ct[j]; // mean
				ct_var[j] = ct_var[j]/(double)ct[j] - ct_mean[j]*ct_mean[j]; // variance
				ct_var[j] = std::min(ct_var[j], 0.01); // max std dev of 0.3 m
				//cout << j << "mean: " << ct_mean[j] << "; variance: " << ct_var[j] << endl;
				planes[j].set(0.0);
			}
			ii = 0;
			ip = P.begin();
			in = N.begin();
			for ( ; ii<n_pts; ++ii, ++ip, ++in ) { // calculate new weight p's
				if ( (asg=ix[ii]) == -1 ) continue;
				gd = exp( -0.5 * (w[ii]-ct_mean[asg])*(w[ii]-ct_mean[asg]) / ct_var[asg] );
				if ( gd < 0.01 ) { // over 3 sigma away: evict point
					ix[ii] = -1;
					--ct[asg];
				}
				w_tot[asg] += gd;
				planes[asg].p += gd * *ip;
				planes[asg].n += gd * *in;
			}
			for ( int j=0; j<k; ++j ) {
				if ( ct[j] > 0 ) {
					planes[j].p /= w_tot[j];
					planes[j].n /= w_tot[j];
					planes[j].off = -planes[j].n.dot(planes[j].p);
				}
			}
		}
#else
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				planes[j].p /= ct[j];
				planes[j].n /= ct[j];
				planes[j].off = -planes[j].n.dot(planes[j].p);
			}
		}
#endif

#if 0 // get n via SVD
		// recalculate centers
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				cvZero(PPTS[j]); // zero matrices
			}
		}
		ip = P.begin();
		ii = 0;
		while ( ii < n_pts ) {
			if ( ix[ii] != -1 ) {
				// accumulate upper-triangular part of matrix
				svlPoint3d pt = *ip - planes[ix[ii]].p;
				CV_MAT_ELEM(*PPTS[ix[ii]],float,0,0) += (float)(pt.x*pt.x);
				CV_MAT_ELEM(*PPTS[ix[ii]],float,0,1) += (float)(pt.x*pt.y);
				CV_MAT_ELEM(*PPTS[ix[ii]],float,0,2) += (float)(pt.x*pt.z);
				CV_MAT_ELEM(*PPTS[ix[ii]],float,1,1) += (float)(pt.y*pt.y);
				CV_MAT_ELEM(*PPTS[ix[ii]],float,1,2) += (float)(pt.y*pt.z);
				CV_MAT_ELEM(*PPTS[ix[ii]],float,2,2) += (float)(pt.z*pt.z);
			}
			++ip;
			++ii;
		}
		// set lower-triangular part of covariance matrix and do SVD
		for ( int j=0; j<k; ++j ) {
			if ( ct[j] > 0 ) {
				CV_MAT_ELEM(*PPTS[j],float,1,0) = CV_MAT_ELEM(*PPTS[j],float,0,1);
				CV_MAT_ELEM(*PPTS[j],float,2,0) = CV_MAT_ELEM(*PPTS[j],float,0,2);
				CV_MAT_ELEM(*PPTS[j],float,2,1) = CV_MAT_ELEM(*PPTS[j],float,1,2);
				cvSVD(PPTS[j],S,U,NULL,0); // SVD
				planes[j].n.x = CV_MAT_ELEM(*U,float,0,2);
				planes[j].n.y = CV_MAT_ELEM(*U,float,1,2);
				planes[j].n.z = CV_MAT_ELEM(*U,float,2,2);
				//cout << "Reassign " << j << " (" << ct[j] << "): " << planes[j].n << endl;
			}
		}
#endif

		// sum deviations from old centers
		gd = 0.0;
		for ( int j=0; j<k; ++j ) {
			if ( planes[j].flag ) {
				if ( ct[j] == 0 ) {
					if ( strcmp(getOption("EmptyAction").c_str(),"Restart") == 0 ) { // restart center at random
						gi = rand()%n_pts;
						planes[j].p = P.at(gi);
						planes[j].n = N.at(gi);
					} else if ( strcmp(getOption("EmptyAction").c_str(),"Drop") == 0 ) { // drop center from rest of algorithm
						planes[j].flag = false; // stop tracking and updating centroid j
					}
				}
				gd += sqrt( (planes[j].p-planesp[j].p).norm2() + (planes[j].n-planesp[j].n).norm2() ); // add distance
			}
		}
		//cout << "Iter # " << iter << " -- Conv: " << gd << endl;
	}

	//int n_max = relabel_asgs(ix); // cannot do this or assignments won't agree with centroids in CENTS
	// release matrices
	for ( int i=0; i<k; ++i )
		cvReleaseMat(&PPTS[i]);
    cvReleaseMat(&U);
    cvReleaseMat(&S);

	/*FILE *fp = fopen("k_n.out","wb");
	fwrite(&n_pts,sizeof(int),1,fp);
	for ( int i=0; i<n_pts; ++i ) {
		fwrite(&P[i].x,sizeof(double),1,fp);
		fwrite(&P[i].y,sizeof(double),1,fp);
		fwrite(&P[i].z,sizeof(double),1,fp);
	}
	for ( int i=0; i<k; ++i ) {
		if ( plane_enabled[i] ) {
			fwrite(&planes[i].p.x,sizeof(double),1,fp);
			fwrite(&planes[i].p.y,sizeof(double),1,fp);
			fwrite(&planes[i].p.z,sizeof(double),1,fp);
			fwrite(&planes[i].n.x,sizeof(double),1,fp);
			fwrite(&planes[i].n.y,sizeof(double),1,fp);
			fwrite(&planes[i].n.z,sizeof(double),1,fp);
		}
	}
	fclose(fp);*/

	// MergeClusters
	if ( getOptionAsBool("Merge") ) {
		for ( int i=0; i<k; ++i ) {
			if ( planes[i].flag ) {
				for ( int j=i+1; j<k; ++j ) {
					if ( planes[j].flag ) {
						svlPoint3d p_offset = planes[i].p - planes[j].p;
						if ( fabs(planes[i].n.dot(planes[j].n)) >= 0.93
							&& fabs(p_offset.dot(planes[i].n)) <= 0.5
							&& fabs(p_offset.dot(planes[j].n)) <= 0.5 ) {
							// merge cluster j into cluster i
							for ( int k=0; k<n_pts; ++k ) {
								if ( ix[k] == j ) {
									ix[k] = i;
								}
							}
							planes[j].flag = false;
						}
					}
				}
			}
		}
	}

	// TODO [PB]: Have relabel_asgs return a permutation vector.
	populateRegionIx(ix,k,rix);

	//int writei;
	//double writed = getOptionAsDouble("% of max # pts");
	//FILE *fp = fopen("km.out","wb");
	//fwrite(&writed,sizeof(double),1,fp);
	//if ( (gd=getOptionAsDouble("% of max # pts")) > 0.0 ) {
	//	// use only major clusters; dropping remainder
	//	svlIxVector<int,int> regionSizes((int)rix.size(),true); // true reverses sort (descending)
	//	for ( int i=0; i<(int)rix.size(); ++i ) {
	//		regionSizes.push_back((int)rix[i].size());
	//	}
	//	regionSizes.sort();
	//	int ick = 0;
	//	// count off valid clusters
	//	for ( ; ick<(int)regionSizes.size(); ++ick ) {
	//		if ( regionSizes.at(ick) < gd*regionSizes.at(0) )
	//			break;
	//	}
	//	// falsilize remaining centers
	//	for ( int j=ick; j<(int)regionSizes.size(); ++j ) {
	//		writei = regionSizes.at(j);
	//		fwrite(&writei,sizeof(int),1,fp);
	//		planes[regionSizes.ix_at(j)] = false;
	//	}
	//	// reassign any points that were before assigned to the newly-falsilized clusters
	//	for ( int j=ick; j<(int)regionSizes.size(); ++j ) {
	//		for ( vector<int>::const_iterator n=rix[regionSizes.ix_at(j)].begin(); n!=rix[regionSizes.ix_at(j)].end(); ++n ) {
	//			// find new min for this point *n
	//			gd = 1e9;
	//			for ( int m=0; m<k; ++m ) {
	//				/*if ( plane_enabled[m] ) {
	//					if ( dists[m*n_pts + i] < gd ) {
	//						gd = dists[m*n_pts + *n];
	//						ix[*n] = m;
	//					}
	//				}	*/
	//			}
	//			// manage rix
	//			rix[ix[*n]].push_back(*n);
	//		}
	//		// delete the now-empty region indices vector for defunct cluster
	//		rix[regionSizes.ix_at(j)].clear();
	//	}
	//}
	//fclose(fp);

	// invert all the off's
	for ( int i=0; i<(int)planes.size(); ++i ) {
		planes[i].off *= -1;
	}

	delete[] ct;
	delete[] w;
	delete[] ct_var;
	delete[] ct_mean;
	delete[] w_tot;
}

